/*
 * SPDX-FileCopyrightText: All Contributors to the PyTango project
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include "common_header.h"

namespace PyDeviceData {

py::object extract(py::object py_self, PyTango::ExtractAs extract_as);

} // namespace PyDeviceData
